class ExternalTool:
    """Class containing various functions for controlling the external tool."""
    def __init__(self) -> None:
        self.request = ""
    
    def set_commandeer(self, enabled):
        """Sets the communication mode of the external tool to either manual or commandeer.
        
        `enabled` is a boolean that determines whether a request to enable commandeer or not should be 
        sent. `True` = enable commandeer, `False` = disable."""
        self.request = "COMMANDEER;" + str(enabled) + "\n"


class Emulation:
    """Class containing various functions for controlling emulation in BizHawk."""
    def __init__(self) -> None:
        self.framecount = -1
        self.request = ""

    def request_framecount(self):
        """Requests for the current framecount from the external tool."""
        self.request = "FRAME;\n"

    def process_responses(self, responses):
        """Updates fields accordingly with the given responses."""
        for response in responses.split("\n"):
            if len(response) > 5 and response[0:5] == "FRAME":
                    self.framecount = int(response[6:])



from birdseye.socketClient import Client
from birdseye.memory import Memory
from birdseye.controllerInput import ControllerInput
from birdseye.emulation import Emulation
from birdseye.externalTool import ExternalTool



class Memory:
    """Class containing various functions for setting controller input in BizHawk."""
    def __init__(self) -> None:
        self.address_list = []
        self.request = ""
        self.received_memory = ""

    def add_address(self, addr):
        """Adds an address for the external tool to return.
        
        `addr` is a hexidecimal value representing the address to read from 
        in the BizHawk emulator's memory."""
        if not addr in self.address_list:
            self.address_list.append(str(addr))
            self.received_memory += str(addr) + ":-1;"

    def add_address_range(self, start, end):
        """Adds a range of addresses from `start` to `end`, both inclusive.

        `start` is a hexidecimal value representing the first address in the range.

        `end` is a hexidecimal value representing the last address in the range.
        
        Precondition: `start` <= `end`."""
        for addr in range(int(start), int(end) + 1):
            self.add_address(addr)
    
    def request_memory(self):
        """Requests for the latest memory data from the external tool."""
        self.request = "MEMORY;" + ";".join(self.address_list) + "\n"
        self.address_list = []

    def get_memory(self) -> list:
        """Gets the latest memory data received from the external tool. 
        
        This will return a list of the latest values received from each address paired 
        with the address it was read from, represented as:

        `"addr:-1"`, if there has been no data received yet from an address

        `"addr:data"`, with both `addr` and `data` in decimal form."""
        return self.received_memory.split(";")
    
    def process_responses(self, responses):
        """Updates fields accordingly with the given responses."""
        for response in responses.split("\n"):
            if len(response) > 6 and response[0:6] == "MEMORY":
                    self.received_memory = response[7:]



class ControllerInput:
    """Class containing various functions for reading memory from BizHawk."""
    def __init__(self) -> None:
        self.request = ""

    def set_controller_input(self, a=False, b=False, up=False, down=False, right=False, left=False):
        """Sets the controller inputs to be executed in the emulator.
        All inputs are set to `False` be default.
        The inputs are executed until a new controller input is sent.

        `a` is the state of the A button

        `b` is the state of the B button

        `up` is the state of the Up button on the control pad

        `down` is the state of the Down button on the control pad

        `right` is the state of the Right button on the control pad

        `left` is the state of the Left button on the control pad"""
        bool_to_string = {False : "false", True : "true"}
        controller_input = bool_to_string[a] + ";" + bool_to_string[b] + ";" + \
                           bool_to_string[up] + ";" + bool_to_string[down] + ";" + \
                           bool_to_string[right] + ";" + bool_to_string[left]
        self.request = "INPUT;" + controller_input + "\n"



import socket


class Client:
    """A socket client used to communicate with the external tool 
    located at a set port and address.

    Use the external tool to get and set the port and
    address the server should listen in to.

    `ip` is the socket address

    `port` is the socket port number"""
    def __init__(self, ip, port):
        self.ip = ip
        self.port = port
        self.client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.connection_status = -1

        self.input_request = ""

    def connect(self):
        """Attempt to connect to the external tool.
        
        Call `Client.is_connected()` to see if the attempt was successful."""
        self.connection_status = self.client.connect_ex((self.ip, self.port))
    
    def close(self):
        """Close socket connection with external tool and send a final message
        notifying the external tool before closing."""
        try:
            self.client.sendall("CLOSE;\n".encode())
        except:
            pass
        self.client.close()
        self.connection_status = -1

    def is_connected(self):
        """Returns true if client is connected to the external tool."""
        return self.connection_status == 0
    
    def send_requests(self, objects: list):
        """Send requests from any birdseye object to the external tool.
        
        Precondition: client is connected to a socket."""
        try:
            self.client.sendall(("".join([obj.request for obj in objects])).encode())
        except:
            self.close()
    
    def get_responses(self) -> str:
        """Receive data collected by the external tool, and update each object in `objects` accordingly.
        
        Precondition: client is connected to a socket."""
        try:
            return self.client.recv(2048).decode()
        except:
            self.close()

        return "";
